/* En nthread-impl.h */

/* Versión alternativa para la Pregunta 2 del examen

	En la auxiliar presencial (2022-2) mencioné que la solución podía tener dataraces,
	pero este comentario fue un "error" ya que olvidé un principio del curso: si la
	solución pasa el diagrama de tiempo entonces es aceptada, y la pauta original cumple
	con el diagrama de	tiempo por lo que está correcta.

	Explicando el comentario de la clase:

	Un datarace puede suceder en el caso que luego de que se retoma la tarea de nOxygen,
	luego de que haya estado en suspensión esperando un hidrógeno. Si justo en ese
	momento llega otra invocación de nOxygen que se roba la CPU, puede robarse los
	Hidrogenos que están en la cola. Para solucionar esto se debe puede garantizar que
	cada thread luego de despertar solo retornará, de esta manera los hidrogenos y
	oxigenos son seteados por el último átomo en ser invocado. Otra opción sería
	asignarle un número de atención a través de una variable global para nOxygen, de
	esta manera garantizar que nunca se anticipe otro nOxygen pero esto puede ser mucho
	más problemático de implementar.

	De esta manera, ante test de esfuerzo más intensivos esta solución siempre
	entregaría la respuesta correcta.
*/

// Añadir campos oxy e hydro para el descriptor del thread
typedef struct
{
	... Oxygen oxy;	 // es usado de dos maneras, un thread de nOxygen pone aquí su
					 // propio átomo para enviarlo como mensaje. Un thread de nHydrogen
					 // recibirá por este campo el oxigeno correspondiente.
	Hydrogen hydro1; // Al igual que en oxy, un thread de nHydro pone aquí su propio
					 // átomo y un thread de nOxygen recibe los hidrógenos por ambos.
	Hydrogen hydro2;
} *nThread;

// Añadir nuevos estados para los threads
typedef enum
{
	... WAIT_HYDRO,
	WAIT_OXY
} State;

/* FIN nthread-impl.h */

NthQueue hydroQ; // = MakeQueue()
NthQueue oxyQ;	 // = MakeQueue()

void nOxygen(Oxygen *oxy, Hydrogen **pH1, Hydrogen **pH2)
{
	// Garantizar exclusión mutua con S_C y E_C en ambas funciones
	START_CRITICAL

	nThread thisTh = nSelf();

	// Caso: hay que esperar por Hidrogenos
	if (nth_queueLength(hydroQ) < 2)
	{

		thisTh->oxy = oxy;
		nth_putBack(oxyQ, thisTh);

		// Suspender tarea si no están los 2 hidrógenos
		suspend(WAIT_HYDRO); // no es relevante cambiar estado
		schedule();
		// se despierta luego de que se completó la molecula
		*pH1 = thisTh->hydro1; // Entregar hidrógenos en *pH1 y *pH2
		*pH2 = thisTh->hydro2;
	}

	// Caso: Listo para retornar una molecula
	else
	{
		nThread t1 = nth_getFront(hydroQ); // Extraer 2 tareas hidrógeno
		nThread t2 = nth_getFront(hydroQ);

		t1->oxy = t2->oxy = oxy; // Depositar oxígeno en tareas hidrógeno

		*pH1 = t1->hydro1; // Entregar hidrógenos en *pH1 y *pH2
		*pH2 = t2->hydro1;

		// Retomar las 2 tareas hidrógeno (no es relevante el orden en que se retoman)
		setReady(t1);
		setReady(t2);
		schedule();
	}

	END_CRITICAL
}

nOxygen *nHydrogen(Hydrogen *h)
{

	START_CRITICAL

	nThread thisTh = nSelf();
	// Registrar hidrógeno para ser extraído por la tarea oxígeno
	thisTh->hydro1 = h;
	// Colocar tarea en hydroQ para suspenderla en (*)

	nth_putBack(hydroQ, thisTh);

	// Retomar una tarea oxígeno si hay 2 hidrógenos en hydroQ
	if (nth_queueLength(hydroQ) >= 2 && !nth_emptyQueue(oxyQ))
	{
		// se han completado los 3 atomos, debemos despertar a los demás threads y
		// setearles los atomos correspondientes
		nTask o = nth_getFront(oxyQ);
		nTask h1 = nth_getFront(hydroQ);
		nTask h2 = nth_getFront(hydroQ); // corresponde a este, pero es necesario
										 // sacarlo de la cola de todas maneras

		// Al thread de nOxygen se le deben entregar dos hidrogenos
		o->hydro1 = h1->hydro1;
		o->hydro2 = h2->hydro1;

		// A los threads de nHydrogen basta con entregarles un átomo de oxígeno
		h1->oxy = o->oxy;
		thisTh->oxy = o->oxy; // thisTh = h2;

		setReady(t);
	}
	else
	{
		// debes esperar a que se completen los 3 atomos
		suspend(WAIT_OXY); // no es relevante cambiar estado
	}
	schedule();				   // (*) Suspende tarea hidrógeno
	Oxygen *oxy = thisTh->oxy; // Rescatar oxígeno y retornarlo

	END_CRITICAL
	return oxy;
}
